import numpy as np
import shutil
import os
from io import StringIO
import math, re, sys, logging
from functools import reduce
import time as tm
import datetime as dt
import casatools
msmd = casatools.msmetadata()
tb = casatools.table()

desc = {'ANTENNA_ID': {'comment': 'ID of antenna in this array',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'FEED_ID': {'comment': 'Feed id',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'INTERVAL': {'comment': 'Interval for which this set of parameters is accurate',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {'QuantumUnits': np.array(['s'], dtype='<U1')},
  'maxlen': 0,
  'option': 0,
  'valueType': 'double'},
 'SPECTRAL_WINDOW_ID': {'comment': 'ID for this spectral window setup',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'TIME': {'comment': 'Midpoint of time for which this set of parameters is accurate',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {'MEASINFO': {'Ref': 'UTC', 'type': 'epoch'},
   'QuantumUnits': np.array(['s'], dtype='<U1')},
  'maxlen': 0,
  'option': 0,
  'valueType': 'double'},
 'TSYS': {'comment': 'System temp. for each of the two receptors',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {'QuantumUnits': np.array(['K'], dtype='<U1')},
  'maxlen': 0,
  'ndim': -1,
  'option': 0,
  'valueType': 'float'},
 '_define_hypercolumn_': {},
 '_keywords_': {},
 '_private_keywords_': {}}
dminfo = {'*1': {'COLUMNS': np.array(['ANTENNA_ID', 'FEED_ID', 'INTERVAL', 'SPECTRAL_WINDOW_ID', 'TIME',
         'TSYS'], dtype='<U18'),
  'NAME': 'StandardStMan',
  'SEQNR': 0,
  'SPEC': {'BUCKETSIZE': 1152,
   'IndexLength': 7486,
   'MaxCacheSize': 2,
   'PERSCACHESIZE': 2},
  'TYPE': 'StandardStMan'}}
desc_gc = {'ANTENNA_ID': {'comment': 'Antenna identifier',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'FEED_ID': {'comment': 'Feed identifier',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'GAIN': {'comment': 'Gain polynomial',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'ndim': -1,
  'option': 0,
  'valueType': 'float'},
 'INTERVAL': {'comment': 'Interval for which this set of parameters is accurate',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {'QuantumUnits': np.array(['s'], dtype='<U1')},
  'maxlen': 0,
  'option': 0,
  'valueType': 'double'},
 'NUM_POLY': {'comment': 'Number of terms in polynomial',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'SENSITIVITY': {'comment': 'Antenna sensitivity',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {'QuantumUnits': np.array(['K/Jy'], dtype='<U4')},
  'maxlen': 0,
  'ndim': -1,
  'option': 0,
  'valueType': 'float'},
 'SPECTRAL_WINDOW_ID': {'comment': 'Spectral window identifier',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'int'},
 'TIME': {'comment': 'Midpoint of time for which this set of parameters is accurate',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {'MEASINFO': {'Ref': 'UTC', 'type': 'epoch'},
   'QuantumUnits': np.array(['s'], dtype='<U1')},
  'maxlen': 0,
  'option': 0,
  'valueType': 'double'},
 'TYPE': {'comment': 'Gain curve type',
  'dataManagerGroup': 'StandardStMan',
  'dataManagerType': 'StandardStMan',
  'keywords': {},
  'maxlen': 0,
  'option': 0,
  'valueType': 'string'},
 '_define_hypercolumn_': {},
 '_keywords_': {},
 '_private_keywords_': {}}
dminfo_gc = {'*1': {'COLUMNS': np.array(['ANTENNA_ID', 'FEED_ID', 'GAIN', 'INTERVAL', 'NUM_POLY',
         'SENSITIVITY', 'SPECTRAL_WINDOW_ID', 'TIME', 'TYPE'], dtype='<U18'),
  'NAME': 'StandardStMan',
  'SEQNR': 0,
  'SPEC': {'BUCKETSIZE': 1920,
   'IndexLength': 142,
   'MaxCacheSize': 2,
   'PERSCACHESIZE': 2},
  'TYPE': 'StandardStMan'}}

#################################################################################################
# Main task
def appendantab(vis=None, outvis=None, antab=None, 
                overwrite=False, append_tsys=True, append_gc=True):

    # disallow writing to the input vis
    if vis == outvis:
        print("Please provide a path for outvis different to the input vis")
        return
    
    # get info from vis
    msmd.open(vis)
    ant_names = msmd.antennastations()
    n_band = len(msmd.bandwidths())
    spws = msmd.spwfordatadesc()
    msmd.close()

    # get start time and end time from ms
    tb.open(vis)
    times = tb.getcol('TIME')
    first_time = times[0]
    last_time = times[-1]
    tb.close()

    # remove existing data copy
    if os.path.exists(outvis) and overwrite:
        print("REMOVING EXISTING COPY")
        shutil.rmtree(outvis)
    elif os.path.exists(outvis) and not overwrite:
        print("File ", outvis, " already exists, set overwrite=True or change outvis name")
        return        
    # run the antab parsing and table filling
    shutil.copytree(vis, outvis)
    antab_interp(vis, outvis, antab, ant_names, n_band, spws,
                    first_time, last_time,
                    append_tsys, append_gc, overwrite)



#################################################################################################
# scanner regex from casavlbitools
def s_err(scanner, error):
    raise RuntimeError("line: %d" % scanner.line_no)

def s_keyword(scanner, token):
    if len(token) > 9 or '.' in token:
        res = ('value', token)
    else:
        res = ('key', token.upper())
    return res

def s_newline(scanner, token):
    scanner.line_no += 1
    return None

def s_quote(scanner, token):
    return ('quote', token[1:-1])

def s_number(scanner, token):
    return ('number', float(token))

def s_angle(scanner, token): # also time
    l = token.split(":")
    ## It's not python to use reduce.
    ## But I neither remember nor care what is.
    val = reduce(lambda acc, x: 60*acc+math.copysign(acc, float(x)), l, 0.0) 
    return ('number', val)

def s_value(scanner, token):
    return ('value', token)

def s_string(scanner, token):
    return ('value', token)

def s_misc(scanner, token):
    logging.debug("Misc token %s at line %d" % (str(token), scanner.line_no))
    return ('misc', token)

def s_comment(scanner, token): # was only used for debugging.
    scanner.line_no += 1
    return ('comment', token)

scanner = re.Scanner([
    ("\!.*\n", s_newline),
    ("[ \t]+", None),
    ("\n", s_newline),
    ("\r\n", s_newline),
    ("=", lambda s, t: ('equal',)),
    ("'[^'\n]*'", s_quote), 
    ("\"[^'\n]*\"", s_quote), # Sigh.  Double quotes used in freq.dat
    ("/", lambda s, t: ('end_chunk',)),
    (",", lambda s, t: ('comma',)),
    ("[+-]?[0-9]+:[0-9]+:[0-9]+(.[0-9]*)?", s_angle),
    ("[+-]?[0-9]+:[0-9]+(.[0-9]*)?", s_angle),
    ("[+-]?[0-9]*\.[0-9]+([Ee][+-][0-9]{1,3})?(?![A-Za-z_0-9()])", s_number),
    ("[+-]?[0-9]+\.?(?![A-Za-z_0-9()])", s_number),
    ## apparently parens and unquoted minus signs are allowed in keywords?
    ("[A-Za-z.0-9]([()A-Za-z_0-9._+-]+)?", s_keyword), 
    (".*", s_misc)
])

def p_all():
    global tok
    chunks = []
    try:
        while True:
            if tok=='EOF':break
            chunks.append(p_chunk())
    except StopIteration:
        pass
    return chunks

def p_chunk():
    global tok
    entries = []
    while tok[0] != 'end_chunk':
        entries.append(p_item())
    logging.debug("p_chunk %s", str(tok))
    tok = next(tokIt)
    return entries

def p_item():
    global tok
    lhs = p_key()
    #if tok[0] == 'key':
        #print(tok)
    if tok==('equal',):
        logging.debug("p_item %s", str(tok))
        tok = next(tokIt)
        rhs = p_rhs()
    elif tok[0] in ['value', 'quote', 'number']:
        rhs = p_rhs()
    else:
        rhs = True # for unitary expressions.
    return (lhs, rhs)

def p_key():
    global tok
    logging.debug("p_key: %s", str(tok))
    if tok[0] == 'key':
        res = tok
        tok = next(tokIt)
    else:
        raise RuntimeError("Expected key token, got %s" % str(tok))
    return res[1]

def p_rhs():
    global tok
    val = p_value()
    rhs = [val]
    while tok == ('comma',):
        logging.debug("p_rhs: %s", str(tok))
        tok = next(tokIt)
        rhs.append(p_value()) # p_value advances tok beyond the value.
    if len(rhs)==1:
        rhs = rhs[0]
    return rhs

def p_value():
    global tok
    if tok[0] not in ['value', 'quote', 'number', 'key']:
        raise RuntimeError("Unexpected RHS token %s" % str(tok))
    val = tok
    logging.debug("p_value: %s", str(val))
    tok = next(tokIt)
    return val[1]

scanner.line_no = 0
#################################################################################################
# data table Classes

class TSysTable:
    def __init__(self):
        self.time = []
        self.time_interval = []
        self.source_id = []
        self.antenna_no = []
        self.array = []
        self.freqid = []
        self.tsys_1 = []
        self.tsys_2 = []
        self.tant = []
        return

class GainCurveTable:
    def __init__(self):
        self.antenna_no = []
        self.array = []
        self.freqid = []
        self.nterm = []
        self.y_typ = []
        self.gain = []
        self.sens_1 = []
        self.sens_2 = []
        return
    
##################################################################


def read_keyfile(f):
    global tok, tokIt
    scanner.line_no = 0

    try:
        res = scanner.scan(f.read())
        #print(res[0])
        if res[1]!='':
            raise RuntimeError("Unparsed text: %s." % (res[1][:20]))
        tokIt = iter(res[0]+['EOF'])
        try: 
            tok = next(tokIt)
            res = p_all()
        except StopIteration: # empty file
            res = ''
    except RuntimeError as txt:
        #print("line %d:  %s" % (scanner.line_no, txt), file=sys.stderr)
        raise RuntimeError
    return res

def update_map(pols, spws, spwmap, index):
    idx = 0
    if not isinstance(index, (list, tuple)):
        index = [index]
        pass
    for labels in index:
        for label in labels.split('|'):
            pol = label[0]
            rng = label[1:].split(':')
            if pol != 'X':
                if not pol in pols:
                    pols.append(pol)
                    pass
                if len(rng) == 1:
                    rng.append(rng[0])
                    pass
                rng = [int(x) - 1 for x in rng]
                for spw in range(rng[0], rng[1] + 1):
                    if not spw in spws:
                        spws.append(spw)
                        pass
                    spwmap[(pol, spw)] = idx
                    continue
                pass
            continue
        idx += 1
        continue
    spws = sorted(spws)
    return

def find_antenna(keys, ignore):
    for key in keys[1:]:
        if not type(key[1]) is bool:
            continue
        if key[0] in ignore:
            continue
        return key[0]
    return None

def skip_values(infp):
    for line in infp:
        if line.startswith('!'):
            continue
        if line.strip().endswith('/'):
            break
        continue
    return

def get_antenna_index(antenna_name, ant_names):
    return ant_names.index(antenna_name)

def mjd_to_date(mjd):
    # Method used in FITSDateUtil
    mjd_epoch = dt.datetime(1858, 11 , 17, 0, 0, 0)

    mjd_int = int(mjd / 86400)
    delta = dt.timedelta(days=mjd_int)
    return mjd_epoch + delta

def mjd_seconds(yy, mm, dd, d=0):
    if (mm < 3):
        yy -= 1
        mm += 12

    dd += d
    b = 0

    if (yy>1582 or (yy==1582 and (mm>10 or (mm==10 and dd >= 15)))): 
        b = np.floor(yy/100.)
        b = 2 - b + int(b/4)
    
    val = np.floor(365.25*yy) + np.floor(30.6001*(mm+1)) + dd - 679006.0 + b
    return val


def get_timetuple(ts):
    # ts as string with these possible formats:
    # hh.hh      
    # hh:mm.mm   
    # hh:mm:ss.ss
    # NOTE: Regexs below will match any number of decimals on the last quantity (e.g. 19.8222222 and 19.8 both work)
    if re.match(r"[0-9]{2}\.[0-9]+", ts):
        # hh.hh 
        tm_hour = int(ts.split('.')[0])
        tm_min = math.modf(60*float(ts.split('.')[1]))
        tm_sec = int(60 * tm_min[0])
        tm_min = int(tm_min[1])
    elif re.match(r"[0-9]{2}:[0-9]{2}\.[0-9]+", ts):
        # hh:mm.mm 
        tm_hour = int(ts.split(':')[0])
        tm_min = math.modf(float(ts.split(':')[1]))
        tm_sec = int(60 * tm_min[0])
        tm_min = int(tm_min[1])
    elif re.match(r"[0-9]{2}:[0-9]{2}:[0-9]{2}$", ts):
        # hh:mm:ss
        tm_hour = int(ts.split(':')[0])
        tm_min = int(ts.split(':')[1])
        tm_sec = int(ts.split(':')[2])
    elif re.match(r"[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]+", ts):
        # hh:mm:ss.ss
        tm_hour = int(ts.split(':')[0])
        tm_min = int(ts.split(':')[1])
        tm_sec = float(ts.split(':')[2])
    return tm_hour, tm_min, tm_sec

def process_tsys(infp, keys, pols, data, n_band, spws, first_time, last_time, ant_names):
    # Get the antenna name
    antenna_name = find_antenna(keys[0], ['SRC/SYS'])
    # Skip if no antenna name found
    if not antenna_name:
        print('ANTENNA missing from TSYS group')
        skip_values(infp)
        return
    try:
        antenna = get_antenna_index(antenna_name, ant_names)
    except:
        print('Antenna %s not present in FITS-IDI file' % antenna_name)
        skip_values(infp)
        return
    
    keys = dict(keys[0])
    spws = []
    spwmap = {}
    
    if 'INDEX' in keys:
        update_map(pols, spws, spwmap, keys['INDEX'])
    if 'INDEX2' in keys:
        update_map(pols, spws, spwmap, keys['INDEX2'])
        pass
    if len(spws) != n_band:
        print('INDEX for antenna %s does not match FITS-IDI file'
              % antenna_name, file=sys.stderr)
        pass

    spws = range(n_band)
    timeoff = 0
    if 'TIMEOFF' in keys:
        timeoff = float(keys['TIMEOFF'])

    for line in infp:
        if line.startswith('!'):
            continue

        fields = line.split()

        if len(fields) > 1:
            tm_yday = int(fields[0])
            # Get timestamp data depending on data format
            obs_date = mjd_to_date(first_time)
            tm_year = obs_date.year
            idi_time = tm.strptime(f"{obs_date.year}-{obs_date.month}-{obs_date.day}","%Y-%m-%d")
            idi_ref = tm.mktime(idi_time)

            # Get the start time in sec for MJD using the method from the table filler
            mjd_sec = mjd_seconds(int(obs_date.year), int(obs_date.month), int(obs_date.day)) * 86400

            tm_hour, tm_min, tm_sec = get_timetuple(fields[1])
            days = (tm_hour*3600 + tm_min*60 + tm_sec) / 86400
            t = "%dy%03dd%02dh%02dm%02ds" % \
                (tm_year, tm_yday, tm_hour, tm_min, tm_sec)
            t = tm.mktime(tm.strptime(t, "%Yy%jd%Hh%Mm%Ss"))
            days = (t + timeoff - idi_ref) / 86400

            curr_time = mjd_sec + (days*86400)

            source = 0
            if (days) > ((last_time-mjd_sec)/86400) or (days) < ((first_time-mjd_sec)/86400):
                source = -1

            #print(deg - mjd_sec)
            values = fields[2:]
            tsys = {'R': [], 'L': []}
            for spw in spws:
                for pol in ['R', 'L']:
                    try:
                        value = float(values[spwmap[(pol, spw)]])
                        if value == 999.9:
                            value = float(-999.9)
                            pass
                    except:
                        value = float(-999.9)
                        pass
                    tsys[pol].append(value)
                    continue
                continue
            if source != -1:
                time_val = mjd_sec + (days * 86400)
                data.time.append(time_val)
                data.time_interval.append(0.0)
                data.antenna_no.append(antenna)
                data.tsys_1.append(tsys['R'])
                data.tsys_2.append(tsys['L'])

                pass
            pass

        if line.strip().endswith('/'):
            break
        continue
    return

gain_keys = [ 'EQUAT', 'ALTAZ', 'ELEV', 'GCNRAO', 'TABLE', 'RCP', 'LCP' ]

def process_gc(infp, keys, pols, data, ant_names):
    antenna_name = find_antenna(keys[0], gain_keys)
    if not antenna_name:
        print('Antenna missing from GAIN group')
        skip_values(infp)
        return
    try:
        antenna = get_antenna_index(antenna_name, ant_names)
    except:
        print('Antenna %s not present in FITS-IDI file' % antenna_name)
        skip_values(infp)
        return
    keys = dict(keys[0])

    dpfu = {}
    try:
        dpfu['R'] = keys['DPFU'][0]
        dpfu['L'] = keys['DPFU'][1]
        pols.append('R')
        pols.append('L')
    except:
        dpfu['R'] = dpfu['L'] = keys['DPFU']
        pols.append('X')
        pass
    try:
        value = keys['POLY'][0]
    except:
        keys['POLY'] = [keys['POLY']]
        pass

    y_typ = 0
    if 'ELEV' in keys:
        y_typ = 1
    elif 'EQUAT' in keys:
        y_typ = 1
    elif 'ALTAZ' in keys:
        y_typ = 2
    else:
        print('Unknown gain curve type for antenna %s' % antenna_name)
        return

    poly = keys['POLY']
    data.antenna_no.append(antenna)
    data.array.append(1)
    data.freqid.append(1)
    data.y_typ.append(y_typ)
    data.nterm.append(len(poly))
    data.gain.append(poly)
    data.sens_1.append(dpfu['R'])
    data.sens_2.append(dpfu['L'])
    return

def antab_interp(vis, outvis, antab, ant_names, n_band, spws,
                    first_time, last_time,
                    append_tsys, append_gc, overwrite):
    pols = []
    data = TSysTable()
    data_gc = GainCurveTable()
    keys = StringIO()
    fp = open(antab, 'r')

    for line in fp:
        if line.startswith('!'):
            continue
        keys.write(line)
        if line.strip().endswith('/'):
            keys.seek(0)
            try:
                tsys = read_keyfile(keys)
            except RuntimeError:
                print("\n", keys.getvalue(), file=sys.stderr)
                raise RuntimeError('error parsing ANTAB file')
            if tsys and tsys[0] and tsys[0][0][0] == 'TSYS':
                process_tsys(fp, tsys, pols, data, n_band, spws, first_time, last_time, ant_names)
                pass
            elif tsys and tsys[0] and tsys[0][0][0] == 'GAIN':
                process_gc(fp, tsys, pols, data_gc, ant_names)
            keys = StringIO()
            continue
        continue
    
    # Create the subtable for syscal if selected in params
    if append_tsys: 
        create_syscal_subtable(data, outvis, spws, overwrite)
    if append_gc: 
        create_gaincurve_subtable(data_gc, outvis, spws, overwrite)

def create_gaincurve_subtable(data, outvis, spws, overwrite):
    # If no subtable exists add one
    if not os.path.exists(f'{outvis}/GAIN_CURVE'):
        # Create the subtable for gain curve
        tb.create(f'{outvis}/GAIN_CURVE', desc_gc, dminfo=dminfo_gc)
        tb.putkeyword('GAIN_CURVE', f'Table: {outvis}/GAIN_CURVE')

    # assemble columns
    ANTENNA_ID = []
    FEED_ID = []
    SPECTRAL_WINDOW_ID = []
    TIME = []
    INTERVAL = []
    TYPE = []
    NUM_POLY = []
    GAIN = []
    SENSITIVITY = [[],[]]

    # If a subtable exists then read contents into assembed data
    # Create data structure for antenna an time stamps to ignore if not overwriting
    existing_data = dict()
    replacements = dict()
    if os.path.exists(f'{outvis}/GAIN_CURVE'):
        tb.open(f'{outvis}/GAIN_CURVE')
        antennas = tb.getcol('ANTENNA_ID')
        times = tb.getcol('TIME')

        # add a set of times for each antenna
        for i in range(len(antennas)):
            if antennas[i] not in existing_data:
                existing_data[antennas[i]] = dict()
                existing_data[antennas[i]]['times'] = set()

                existing_data[antennas[i]]['times'].add(times[i])
                existing_data[antennas[i]]['row'] = i
            else:
                existing_data[antennas[i]]['times'].add(times[i])
                existing_data[antennas[i]]['row'] = i
                
        tb.close()

    # Get interval information from the observation table instead of meta-data
    tb.open(f'{outvis}/OBSERVATION')
    time_range = tb.getcol("TIME_RANGE")
    gc_interval = (time_range[1] - time_range[0])[0]
    gc_time = np.mean(time_range)
    tb.close()

    # Fill the arrays
    for i in range(len(data.antenna_no)):
        for j in spws:
            # if the antenna has been visited already with the same time stamp then ignore
            if data.antenna_no[i] in existing_data and gc_time in existing_data[data.antenna_no[i]] and not overwrite:
                # This data is already in the subtable so ignore
                #print("antenna: {} has data present for the time {}. The antab data for this value will be ignored".format(data.antenna_no[i], gc_time))
                continue
            elif data.antenna_no[i] in existing_data and gc_time in existing_data[data.antenna_no[i]] and overwrite:
                # Get the row and information to overwrite the row from the ms
                row = existing_data[data.antenna_no[i]]['row']

                replacements[row] = {'ANTENNA_ID':data.antenna_no[i], 'FEED_ID':-1, 'INTERVAL':gc_interval,
                                    'TYPE':data.y_typ[i], 'NUM_POLY':data.nterm[i], 'GAIN':data.gain[i],
                                    'TIME':gc_time, 'SPECTRAL_WINDOW_ID':int(j), 'SENSITIVITY':[[data.sens_1[i]], [data.sens_2[i]]]}

                continue

            TIME.append(gc_time)
            ANTENNA_ID.append(data.antenna_no[i])
            FEED_ID.append(-1)
            INTERVAL.append(gc_interval)
            TYPE.append(str(data.y_typ[i]))
            NUM_POLY.append(data.nterm[i])
            GAIN.append(data.gain[i])
            SPECTRAL_WINDOW_ID.append(int(j))
            SENSITIVITY[0].append(data.sens_1[i])
            SENSITIVITY[1].append(data.sens_2[i])

    # If appending get existing cols and append values to it
    tb.open(f'{outvis}/GAIN_CURVE', nomodify=False)
    if not overwrite and tb.nrows() > 0:
        #tb.addrows(len(TIME))
        TIME = np.append(tb.getcol('TIME'), TIME)
        ANTENNA_ID = np.append(tb.getcol('ANTENNA_ID'), ANTENNA_ID)
        FEED_ID = np.append(tb.getcol('FEED_ID'), FEED_ID)
        INTERVAL = np.append(tb.getcol('INTERVAL'), INTERVAL)
        TYPE = np.append(tb.getcol('TYPE'), TYPE)
        NUM_POLY = np.append(tb.getcol('NUM_POLY'), NUM_POLY)

        # Can't call getcol on Gain since dimensions can vary
        tmp_GAIN = []
        for row in range(tb.nrows()):
            tmp_GAIN.append(tb.getcell('GAIN', row))
        # re-assembled the old and new gain col
        GAIN = tmp_GAIN + GAIN
        SPECTRAL_WINDOW_ID = np.append(tb.getcol('SPECTRAL_WINDOW_ID'), SPECTRAL_WINDOW_ID)
        tmp_0 = np.append(tb.getcol('SENSITIVITY')[0], SENSITIVITY[0])
        tmp_1 = np.append(tb.getcol('SENSITIVITY')[1], SENSITIVITY[1])
        SENSITIVITY = np.asarray([tmp_0, tmp_1])

    # If overwriting then go over the existing rows to overwrite
    if overwrite:
        for row, data in replacements.items():
            TIME[row] = items['TIME']
            ANTENNA_ID[row] = items['ANTENNA']
            FEED_ID[row] = items['FEED_ID']
            INTERVAL[row] = items['INTERVAL']
            TYPE[row] = items['TYPE']
            NUM_POLY[row] = items['NUM_POLY']
            GAIN[row] = items['GAIN']
            SPECTRAL_WINDOW_ID[row] = items['SPECTRAL_WINDOW_ID']
            # HOW TO ASSEMBLE SENSITIVITY?
            SENSITIVITY[0][row] = items['SENSITIVITY'][0]
            SENSITIVITY[1][row] = items['SENSITIVITY'][1]
            
    tb.close()

    # Add the columns to the table
    tb.open(f'{outvis}/GAIN_CURVE', nomodify=False)
    to_add = len(TIME) - tb.nrows()
    tb.addrows(to_add)
    tb.putcol('TIME', TIME)
    tb.putcol('ANTENNA_ID', ANTENNA_ID)
    tb.putcol('FEED_ID', FEED_ID)
    tb.putcol('SPECTRAL_WINDOW_ID', SPECTRAL_WINDOW_ID)
    tb.putcol('INTERVAL', INTERVAL)
    # Type 1 POWER(EL) Type 2 POWER(ZA)
    tb.putcol('NUM_POLY', NUM_POLY)
    #Fill gain row by row
    for i in range(len(GAIN)):
        # If the gain col shape doesn't match what the table filler does
        if len(np.shape(GAIN[i])) < 2:
            tb.putcell('GAIN', i, [GAIN[i],GAIN[i]])
        else:
            tb.putcell('GAIN', i, GAIN[i])
        
        # Convert Type to proper string
        if TYPE[i] == '1':
            TYPE[i] = 'POWER(EL)'
        elif TYPE[i] == '2':
            TYPE[i] = 'POWER(ZA)'

    tb.putcol('TYPE', TYPE)
    tb.putcol('SENSITIVITY', SENSITIVITY)
    #tb.flush()
    tb.close()


def create_syscal_subtable(data, outvis, spws, overwrite):    
    # If no subtable exists
    if not os.path.exists(f'{outvis}/SYSCAL'):
        # Create the subtable for syscal
        tb.create(f'{outvis}/SYSCAL', desc, dminfo=dminfo)
        tb.putkeyword('SYSCAL', f'Table: {outvis}/SYSCAL')

    # Assemble columns
    ANTENNA_ID = []
    FEED_ID = []
    INTERVAL = []
    SPW_ID = []
    TIME = []
    TSYS = [[],[]]
    data.tsys_1 = np.asarray(data.tsys_1)
    data.tsys_2 = np.asarray(data.tsys_2)

    # If a subtable exists then read contents into assembed data
    # Create data structure for antenna an time stamps to ignore if not overwriting
    existing_data = dict()
    if os.path.exists(f'{outvis}/SYSCAL'):
        tb.open(f'{outvis}/SYSCAL')
        antennas = tb.getcol('ANTENNA_ID')
        times = tb.getcol('TIME')

        # add a set of times for each antenna
        for i in range(len(antennas)):
            if antennas[i] not in existing_data:
                existing_data[antennas[i]] = dict()
                existing_data[antennas[i]]['times'] = set()

                existing_data[antennas[i]]['times'].add(times[i])
                existing_data[antennas[i]]['row'] = i
            else:
                existing_data[antennas[i]]['times'].add(times[i])
                existing_data[antennas[i]]['row'] = i
                
        tb.close()

    # Fill the arrays
    to_skip = []
    replacements = dict()
    for i in range(len(data.time)):
        # if the antenna has been visited already with the same time stamp then ignore
        if data.antenna_no[i] in existing_data and data.time[i] in existing_data[data.antenna_no[i]] and not overwrite:
            # This data is already in the subtable so ignore
            #print("antenna: {} has data present for the time {}. The antab data for this value will be ignored".format(data.antenna_no[i], data.time[i]))
            to_skip.append(True)
            continue
        else:
            to_skip.append(False)

        for j in spws:
            if data.antenna_no[i] in existing_data and data.time[i] in existing_data[data.antenna_no[i]] and overwrite:
                # Get the row and information to overwrite the row from the ms
                row = existing_data[data.antenna_no[i]]['row']
                to_skip.append(True)

                replacments[row] = {'ANTENNA':data.antenna_no[i], 'FEED_ID':0, 'INTERVAL':data.time_interval[i],
                                    'SPW_ID':int(j), 'TIME':data.time[i], 'TSYS':[[],[]]}

                continue

            ANTENNA_ID.append(data.antenna_no[i])
            FEED_ID.append(0)
            INTERVAL.append(data.time_interval[i])
            SPW_ID.append(int(j))
            TIME.append(data.time[i])

    for i in range(len(data.tsys_1)):
        if to_skip[i] and not overwrite:
            continue
        elif to_skip[i] and overwrite:
            replacements[i]['TSYS'][0].extend(data.tsys_1[i])
            replacements[i]['TSYS'][1].extend(data.tsys_2[i])
            continue
        TSYS[0].extend(data.tsys_1[i])
        TSYS[1].extend(data.tsys_2[i])

    # If appending get existing cols and append values to it also check if empty
    tb.open(f'{outvis}/SYSCAL', nomodify=False)
    if not overwrite and tb.nrows() > 0:
        TIME = np.append(tb.getcol('TIME'), TIME)
        ANTENNA_ID = np.append(tb.getcol('ANTENNA_ID'), ANTENNA_ID)
        INTERVAL = np.append(tb.getcol('INTERVAL'), INTERVAL)
        SPW_ID = np.append(tb.getcol('SPECTRAL_WINDOW_ID'), SPW_ID)
        tmp_0 = np.append(tb.getcol('TSYS')[0], TSYS[0])
        tmp_1 = np.append(tb.getcol('TSYS')[1], TSYS[1])
        TSYS = np.asarray([tmp_0, tmp_1])

    # If overwriting then go over the existing rows to overwrite
    if overwrite:
        for row, data in replacements.items():
            TIME[row] = items['TIME']
            ANTENNA_ID[row] = items['ANTENNA']
            INTERVAL[row] = items['INTERVAL']
            SPW_ID[row] = items['SPW_ID']
            print(TSYS[0][row], TSYS[1][row])
            TSYS[0][row] = items['TSYS'][0]
            TSYS[1][row] = itmes['TSYS'][1]
    tb.close()

    # Add the columns to the table
    tb.open(f'{outvis}/SYSCAL', nomodify=False)
    to_add = len(TIME) - tb.nrows()
    tb.addrows(to_add)
    tb.putcol('TIME', TIME)
    tb.putcol('ANTENNA_ID', ANTENNA_ID)
    tb.putcol('INTERVAL', INTERVAL)
    tb.putcol('SPECTRAL_WINDOW_ID', SPW_ID)
    tb.putcol('TSYS', TSYS)
    #tb.flush()
    tb.close()
