##################### generated by xml-casa (v2) from simalma.xml ###################
##################### 4a3ac82c1288b5699e222ce63d959cad ##############################
from __future__ import absolute_import
import numpy
from casatools.typecheck import CasaValidator as _val_ctor
_pc = _val_ctor( )
from casatools.coercetype import coerce as _coerce
from casatools.errors import create_error_string
from .private.task_simalma import simalma as _simalma_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log
from casatasks.private.task_logging import except_log as _except_log

class _simalma:
    """
    simalma ---- Simulation task for ALMA

    
    This task simulates ALMA observation including 12-m, ACA 7-m and total
    power arrays, and images and analyzes simulated data.
    
    This task makes multiple calls to simobserve (to calculate
    visibilities and total power spectra), followed by gridding of total
    power spectra (if total power is requested), concatenation of the
    simulated visibilities, calls to the simanalyze task for visibility
    inversion and deconvolution and calculation of difference and fidelity
    images, and feathering of single dish and interferometric data.
    
    These steps may not all be familiar to new users, so the simalma task
    runs by default in a "dryrun" mode, in which it assesses the user's
    input parameters and sky model, and prints an informational report
    including the required calls to other CASA tasks, both to the screen
    and to a text file in the project directory (defined below).
    
    The user can modify their parameters based on the information, then
    either run with dryrun=False to actually call the other tasks to
    create the simulated data, or run the other tasks individually one at
    a time to better understand and control the process.
    
    NOTE The ALMA project is refining the optimal method of combining the
    three types of data.  If that best practice is changed after this
    release of CASA, the user can control the process by modifying the
    calls to the other CASA tasks.

    --------- parameter descriptions ---------------------------------------------

    project      root prefix for output file names
    dryrun       dryrun=True will only produce the informative report, not run simobserve/analyze
    skymodel     model image to observe
    inbright     scale surface brightness of brightest pixel e.g. "1.2Jy/pixel"
    indirection  set new direction e.g. "J2000 19h00m00 -40d00m00"
    incell       set new cell/pixel size e.g. "0.1arcsec"
    incenter     set new frequency of center channel e.g. "89GHz" (required even for 2D model)
    inwidth      set new channel width e.g. "10MHz" (required even for 2D model)
    complist     componentlist to observe
    compwidth    bandwidth of components
    setpointings If True, calculate a map of pointings and write *ptgfile*.
    ptgfile      list of pointing positions
    integration  integration (sampling) time
    direction    "J2000 19h00m00 -40d00m00" or "" to center on model
    mapsize      angular size of map or "" to cover model
    antennalist  antenna position files of ALMA 12m and 7m arrays
    hourangle    hour angle of observation center e.g. -3:00:00, or "transit"
    totaltime    total time of observation; vector corresponding to antennalist
    tpnant       Number of total power antennas to use (0-4)
    tptime       total observation time for total power
    pwv          Precipitable Water Vapor in mm. 0 for noise-free simulation
    image        image simulated data
    imsize       output image size in pixels (x,y) or 0 to match model
    imdirection  set output image direction, (otherwise center on the model)
    cell         cell size with units or "" to equal model
    niter        maximum number of iterations (0 for dirty image)
    threshold    flux level (+units) to stop cleaning
    graphics     display graphics at each stage to [screen|file|both|none]
    verbose      Print extra information to the logger and terminal.
    overwrite    overwrite files starting with $project
    [1;42mRETURNS[1;m         bool

    --------- examples -----------------------------------------------------------

    
    -------------------------------
    Parameters:
    
    project -- root filename for all output files.  A subdirectory will be
    created, and all created files will be placed in that subdirectory
    including the informational report.
    
    -------------------------------
    skymodel -- input image (used as a model of the sky)
    * simalma requires a CASA or fits image. If you merely have a grid of
    numbers, you will need to write them out as fits or write a
    CASA script to read them in and use the ia tool to create an image
    and insert the data.
    
    * simalma does NOT require a coordinate system in the header. If the
    coordinate information is incomplete, missing, or you would like to
    override it, set the appropriate "in" parameters. NOTE that setting
    those parameters simply changes the header values, ignoring
    any values already in the image. No regridding is performed.
    
    * If you have a proper Coordinate System, simalma will do its best to
    generate visibilities from that, and then create a synthesis image
    according to the specified user parameters.
    
    * You can manipulate an image header manually with the "imhead" task.
    
    inbright -- peak brightness to scale the image to in Jy/pixel,
    or "" for unchanged
    * NOTE: "unchanged" will take the numerical values in your image
    and assume they are in Jy/pixel, even if it says some other unit
    in the header.
    
    indirection -- central direction to place the sky model image,
    or "" to use whatever is in the image already
    
    incell -- spatial pixel size to scale the skymodel image,
    or "" to use whatever is in the image already.
    
    incenter -- frequency to use for the center channel (or only channel,
    if the skymodel is 2D)  e.g. "89GHz",
    or "" to use what is in the header.
    
    inwidth -- width of channels to use, or "" to use what is in the image
    should be a string representing a quantity with units e.g. "10MHz"
    * NOTE: only works reliably with frequencies, not velocities
    * NOTE: it is not possible to change the number of spectral planes
    of the sky model, only to relabel them with different frequencies
    That kind of regridding can be accomplished with the CASA toolkit.
    
    -------------------------------
    complist -- component list model of the sky, added to or instead of skymodel
    see http://casaguides.nrao.edu/index.php?title=Simulation_Guide_Component_Lists_%28CASA_4.1%29
    
    compwidth -- bandwidth of components; if simulating from components only,
    this defines the bandwidth of the MS and output images
    
    -------------------------------
    setpointings -- if true, calculate a map of pointings and write ptgfile.
    * if graphics are on, display the pointings shown on the model image
    * observations with the ALMA 12m and ACA 7m arrays will observe a
    region of size "mapsize" using the same hexagonal algorithm as
    the ALMA OT, with Nyquist sampling.
    * The total power array maps a slightly (+1 primary beam) larger area
    than the 12m array does, to improve later image combination.
    It samples the region with lattice grids of spacing 0.33 lambda/D.
    * if setpointings=false, read pointings from ptgfile.
    
    ptgfile -- a text file specifying directions in the same
    format as the example, and optional integration times, e.g.
    #Epoch     RA          DEC      TIME(optional)
    J2000 23h59m28.10 -019d52m12.35 10.0
    * if the time column is not present in the file, it will use
    "integration" for all pointings.
    * NOTE: at this time the file should contain only science pointings:
    simalma will observe these until totaltime is used up.
    
    integration --- Time interval for each integration e.g '10s'
    * NOTE: to simulate a "scan" longer than one integration, use
    setpointings to generate a pointing file, and then edit the
    file to increase the time at each point to be larger than
    the parameter integration time.
    
    direction -- mosaic center direction e.g 'J2000 19h00m00 -40d00m00'
    if unset, will use the center of the skymodel image.
    * can optionally be a list of pointings, otherwise
    * simobserve will cover a region of size mapsize according to maptype
    
    mapsize -- angular size of mosaic map to simulate.
    * set to "" to cover the model image
    
    -------------------------------
    antennalist -- vector of ascii files containing antenna positions,
    one for each configuration of 7m or 12m dishes.
    * NOTE: In this task, it should be an ALMA configuration.
    * standard arrays are found in your CASA data repository,
    os.getenv("CASAPATH").split()[0]+"/data/alma/simmos/"
    * a string of the form "alma;0.5arcsec" will be parsed into a
    12m ALMA configuration - see casaguides.nrao.edu
    * examples: ['alma.cycle2.5.cfg','aca.cycle2.i.cfg']
    ['alma.cycle1;0.3arcsec','alma.cycle1.1.cfg','aca.i.cfg']
    
    hourangle -- hour angle of observation e.g. '-3h'
    
    totaltime --- total time of observations. This should either be a scalar
    time quantity expressed as a string e.g. '1h', '3600sec', '10min',
    or a vector of such quantities, corresponding to the elements of
    the antennalist vector, e.g. ['5min','20min','3h'].  If you
    specify a scalar, that will be used for the highest resolution
    12m configuration in antennalist, and any lower resolution 12m
    configurations, any 7m configurations, and any TP configurations
    will have observing times relative to totaltime of 0.5, 2,and 4,
    respectively.
    -------------------------------
    
    tpnant -- the number of total power antennas to use in simulation.
    
    tptime -- if tpnant>0, the user must specify the observing time for
    total power as a CASA quantity e.g. '4h'.
    * NOTE: this is not broken up among multiple days -
    a 20h track will include observations below the horizon,
    which is probably not what is desired.
    
    -------------------------------
    pwv -- precipitable water vapor if constructing an atmospheric model.
    Set 0 for noise-free simulation. When pwv>0, thermal noise is
    applied to the simulated data.
    * J. Pardo's ATM library will be used to construct anatmospheric
    profile for the ALMA site:
    altitude 5000m, ground pressure 650mbar, relhum=20%,
    a water layer of pwv at altitude of 2km,
    the sky brightness temperature returned by ATM, and internally
    tabulated receiver temperatures.
    See the documentation of simobserve for more details.
    
    -------------------------------
    image -- option to invert and deconvolve the simulated measurement set(s)
    * NOTE: interactive clean or more parameters than the subset visible
    here are available by simply running the clean task directly.
    * if graphics turned on, display the clean image and residual image
    * uses Cotton-Schwab clean for single fields and Mosaic gridding
    for multiple fields (with Clark PSF calculation in minor cycles).
    
    imsize -- image size in spatial pixels (x,y)
    0 or -1 will use the model image size; example: imsize=[500,500]
    
    imdirection -- phase center for synthesized image.  default is to
    center on the sky model.
    
    cell -- cell size e.g '10arcsec'.  "" defaults to the skymodel cell
    
    niter -- number of clean/deconvolution iterations, 0 for no cleaning
    
    threshold -- flux level to stop cleaning
    -------------------------------
    graphics -- view plots on the screen, saved to file, both, or neither
    verbose -- print extra information to the logger and terminal
    overwrite -- overwrite existing files in the project subdirectory
    -------------------------------
    
    Please see the documents of simobserve and simanalyze for
    the list of outputs produced.
    


    """

    _info_group_ = """simulation"""
    _info_desc_ = """Simulation task for ALMA """

    def __call__( self, project='sim', dryrun=True, skymodel='', inbright='', indirection='', incell='', incenter='', inwidth='', complist='', compwidth='"8GHz"', setpointings=True, ptgfile='$project.ptg.txt', integration='10s', direction=[  ], mapsize=[ '','' ], antennalist=[ 'alma.cycle1.1.cfg','aca.cycle1.cfg' ], hourangle='transit', totaltime=[ '20min','1h' ], tpnant=int(0), tptime='0s', pwv=float(0.5), image=True, imsize=[ int(128),int(128) ], imdirection='', cell='', niter=int(0), threshold='0.1mJy', graphics='both', verbose=False, overwrite=False ):
        schema = {'project': {'type': 'cStr', 'coerce': _coerce.to_str}, 'dryrun': {'type': 'cBool'}, 'skymodel': {'type': 'cStr', 'coerce': _coerce.to_str}, 'inbright': {'type': 'cStr', 'coerce': _coerce.to_str}, 'indirection': {'type': 'cStr', 'coerce': _coerce.to_str}, 'incell': {'type': 'cStr', 'coerce': _coerce.to_str}, 'incenter': {'type': 'cStr', 'coerce': _coerce.to_str}, 'inwidth': {'type': 'cStr', 'coerce': _coerce.to_str}, 'complist': {'type': 'cStr', 'coerce': _coerce.to_str}, 'compwidth': {'type': 'cStr', 'coerce': _coerce.to_str}, 'setpointings': {'type': 'cBool'}, 'ptgfile': {'type': 'cStr', 'coerce': _coerce.to_str}, 'integration': {'type': 'cStr', 'coerce': _coerce.to_str}, 'direction': {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, 'mapsize': {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, 'antennalist': {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, 'hourangle': {'type': 'cStr', 'coerce': _coerce.to_str}, 'totaltime': {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, 'tpnant': {'type': 'cInt', 'min': 0, 'max': 4}, 'tptime': {'type': 'cStr', 'coerce': _coerce.to_str}, 'pwv': {'type': 'cFloat', 'coerce': _coerce.to_float, 'min': 0}, 'image': {'type': 'cBool'}, 'imsize': {'type': 'cIntVec', 'coerce': [_coerce.to_list,_coerce.to_intvec]}, 'imdirection': {'type': 'cStr', 'coerce': _coerce.to_str}, 'cell': {'type': 'cStr', 'coerce': _coerce.to_str}, 'niter': {'type': 'cInt'}, 'threshold': {'type': 'cStr', 'coerce': _coerce.to_str}, 'graphics': {'type': 'cStr', 'coerce': _coerce.to_str, 'allowed': [ 'none', 'file', 'both', 'screen', '' ]}, 'verbose': {'type': 'cBool'}, 'overwrite': {'type': 'cBool'}}
        doc = {'project': project, 'dryrun': dryrun, 'skymodel': skymodel, 'inbright': inbright, 'indirection': indirection, 'incell': incell, 'incenter': incenter, 'inwidth': inwidth, 'complist': complist, 'compwidth': compwidth, 'setpointings': setpointings, 'ptgfile': ptgfile, 'integration': integration, 'direction': direction, 'mapsize': mapsize, 'antennalist': antennalist, 'hourangle': hourangle, 'totaltime': totaltime, 'tpnant': tpnant, 'tptime': tptime, 'pwv': pwv, 'image': image, 'imsize': imsize, 'imdirection': imdirection, 'cell': cell, 'niter': niter, 'threshold': threshold, 'graphics': graphics, 'verbose': verbose, 'overwrite': overwrite}
        assert _pc.validate(doc,schema), create_error_string(_pc.errors)
        _logging_state_ = _start_log( 'simalma', [ 'project=' + repr(_pc.document['project']), 'dryrun=' + repr(_pc.document['dryrun']), 'skymodel=' + repr(_pc.document['skymodel']), 'inbright=' + repr(_pc.document['inbright']), 'indirection=' + repr(_pc.document['indirection']), 'incell=' + repr(_pc.document['incell']), 'incenter=' + repr(_pc.document['incenter']), 'inwidth=' + repr(_pc.document['inwidth']), 'complist=' + repr(_pc.document['complist']), 'compwidth=' + repr(_pc.document['compwidth']), 'setpointings=' + repr(_pc.document['setpointings']), 'ptgfile=' + repr(_pc.document['ptgfile']), 'integration=' + repr(_pc.document['integration']), 'direction=' + repr(_pc.document['direction']), 'mapsize=' + repr(_pc.document['mapsize']), 'antennalist=' + repr(_pc.document['antennalist']), 'hourangle=' + repr(_pc.document['hourangle']), 'totaltime=' + repr(_pc.document['totaltime']), 'tpnant=' + repr(_pc.document['tpnant']), 'tptime=' + repr(_pc.document['tptime']), 'pwv=' + repr(_pc.document['pwv']), 'image=' + repr(_pc.document['image']), 'imsize=' + repr(_pc.document['imsize']), 'imdirection=' + repr(_pc.document['imdirection']), 'cell=' + repr(_pc.document['cell']), 'niter=' + repr(_pc.document['niter']), 'threshold=' + repr(_pc.document['threshold']), 'graphics=' + repr(_pc.document['graphics']), 'verbose=' + repr(_pc.document['verbose']), 'overwrite=' + repr(_pc.document['overwrite']) ] )
        task_result = None
        try:
            task_result = _simalma_t( _pc.document['project'], _pc.document['dryrun'], _pc.document['skymodel'], _pc.document['inbright'], _pc.document['indirection'], _pc.document['incell'], _pc.document['incenter'], _pc.document['inwidth'], _pc.document['complist'], _pc.document['compwidth'], _pc.document['setpointings'], _pc.document['ptgfile'], _pc.document['integration'], _pc.document['direction'], _pc.document['mapsize'], _pc.document['antennalist'], _pc.document['hourangle'], _pc.document['totaltime'], _pc.document['tpnant'], _pc.document['tptime'], _pc.document['pwv'], _pc.document['image'], _pc.document['imsize'], _pc.document['imdirection'], _pc.document['cell'], _pc.document['niter'], _pc.document['threshold'], _pc.document['graphics'], _pc.document['verbose'], _pc.document['overwrite'] )
        except Exception as exc:
            _except_log('simalma', exc)
            raise
        finally:
            task_result = _end_log( _logging_state_, 'simalma', task_result )
        return task_result

simalma = _simalma( )

